# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

import textwrap

import pytest

from python.cryptopolicies.cryptopolicies import (
    UnscopedCryptoPolicy,
    PolicySyntaxDeprecationWarning,
)
from python.cryptopolicies.alg_lists import glob

from python.cryptopolicies.validation import (
    PolicySyntaxError, PolicyFileNotFoundError
)


TESTPOL = '''
# boring policy
cipher = AES-*-GCM
'''
MINUS192 = 'cipher = -AES-192-*'


def _policy(tmpdir, **kwargs):
    subpolicy = False
    for k, v in kwargs.items():
        if not subpolicy:
            tmpdir.join(f'{k}.pol').write(v)
            subpolicy = True
        else:
            if not tmpdir.join('modules').check(dir=True):
                tmpdir.mkdir('modules')
            tmpdir.join('modules').join(f'{k}.pmod').write(v)
    # pylint: disable=no-value-for-parameter
    return UnscopedCryptoPolicy(*kwargs.keys(), policydir=str(tmpdir))


def test_cryptopolicy_is_empty(tmpdir):
    assert _policy(tmpdir, TESTPOL='').is_empty()
    assert _policy(tmpdir, TESTPOL=' \n\t').is_empty()


def test_cryptopolicy_not_found():
    with pytest.raises(PolicyFileNotFoundError):
        UnscopedCryptoPolicy('NONEX')


def test_cryptopolicy_smoke_broken(tmpdir):
    with pytest.raises(PolicySyntaxError):
        with pytest.warns(PolicySyntaxError):
            _policy(tmpdir, TESTPOL='a = b = c')


def test_cryptopolicy_smoke_basic(tmpdir):
    cp = _policy(tmpdir, TESTPOL='cipher = AES-*-GCM')
    assert cp.scoped({'tls'}).enabled['cipher'] == [
        'AES-256-GCM', 'AES-192-GCM', 'AES-128-GCM'
    ]


def test_cryptopolicy_smoke_subpolicy(tmpdir):
    cp = _policy(tmpdir,
                 TESTPOL='cipher = AES-*-GCM',
                 MINUS192='cipher = -AES-192-*')
    assert cp.scoped({'tls'}).enabled['cipher'] == [
        'AES-256-GCM', 'AES-128-GCM'
    ]


def test_cryptopolicy_smoke_several_subpolicies(tmpdir):
    cp = _policy(tmpdir,
                 TESTPOL='cipher = AES-*-GCM',
                 MINUS_192='cipher = -AES-192-*',
                 TLS_RESET='cipher@TLS = AES-*-GCM',
                 MINUS_128_SSH='cipher@SSH = -AES-128-*',
                 APPEND_NULL_TLS='cipher@TLS = NULL+',
                 PREPEND_RC4='cipher = +RC4-128')
    assert cp.scoped({'tls', 'openssl'}).enabled['cipher'] == [
        'RC4-128', 'AES-256-GCM', 'AES-192-GCM', 'AES-128-GCM', 'NULL'
    ]
    assert cp.scoped({'ssh', 'openssh'}).enabled['cipher'] == [
        'RC4-128', 'AES-256-GCM',
    ]
    assert 'AES-192-GCM' in cp.scoped({'ssh', 'openssh'}).disabled['cipher']
    assert 'AES-128-GCM' in cp.scoped({'ssh', 'openssh'}).disabled['cipher']
    assert 'NULL' in cp.scoped({'ssh', 'openssh'}).disabled['cipher']


def test_cryptopolicy_compat_diamond_new_recommended(tmpdir):
    with pytest.warns(PolicySyntaxDeprecationWarning):
        cp = _policy(tmpdir,
                     TESTPOL='''
                         tls_cipher = DES-CBC RC4-128
                         cipher = RC4-128 IDEA-CBC
                         # ssh_cipher derived as RC4-128 IDEA-CBC
                     ''',
                     TESTSUBPOL='''
                         # as simple and intuitive as it gets
                         cipher = -RC4-128
                     ''')
    assert cp.scoped({'tls', 'openssl'}).enabled['cipher'] == ['DES-CBC']
    assert cp.scoped({'ssh', 'openssh'}).enabled['cipher'] == ['IDEA-CBC']
    assert cp.scoped().enabled['cipher'] == ['IDEA-CBC']


def test_cryptopolicy_compat_diamond_old_recommended(tmpdir):
    with pytest.warns(PolicySyntaxDeprecationWarning):
        cp = _policy(tmpdir,
                     TESTPOL='''
                         tls_cipher = DES-CBC RC4-128
                         cipher = RC4-128 IDEA-CBC
                         # ssh_cipher derived as RC4-128 IDEA-CBC
                     ''',
                     TESTSUBPOL='''
                         # the current 'correct' way to disable ciphers
                         tls_cipher = -RC4-128
                         cipher = -RC4-128
                         ssh_cipher = -RC4-128
                     ''')
    assert cp.scoped({'tls', 'openssl'}).enabled['cipher'] == ['DES-CBC']
    assert cp.scoped({'ssh', 'openssh'}).enabled['cipher'] == ['IDEA-CBC']
    assert cp.scoped().enabled['cipher'] == ['IDEA-CBC']


def test_cryptopolicy_compat_diamond_breaking1(tmpdir):
    with pytest.warns(PolicySyntaxDeprecationWarning):
        cp = _policy(tmpdir,
                     TESTPOL='''
                         tls_cipher = DES-CBC RC4-128
                         cipher = RC4-128 IDEA-CBC
                         # ssh_cipher derived as RC4-128 IDEA-CBC
                     ''',
                     TESTSUBPOL='''
                         # BEHAVIOUR CHANGE!
                         # Modifying cipher in subpolicy previously
                         # didn't affect tls_cipher / ssh_cipher!
                         # Now it does affect cipher@tls / cipher@ssh,
                         # which is a sane, but backwards
                         # incompatible thing to do.
                         cipher = -RC4-128
                     ''')
    # Used to be ['DES-CBC', 'RC4-128']
    assert cp.scoped({'tls', 'openssl'}).enabled['cipher'] == ['DES-CBC']
    # Used to be ['RC4-128', 'IDEA-CBC']
    assert cp.scoped({'ssh', 'openssh'}).enabled['cipher'] == ['IDEA-CBC']
    assert cp.scoped().enabled['cipher'] == ['IDEA-CBC']


def test_cryptopolicy_compat_diamond_breaking2(tmpdir):
    with pytest.warns(PolicySyntaxDeprecationWarning):
        cp = _policy(tmpdir,
                     TESTPOL='''
                         cipher = RC4-128 IDEA-CBC
                         # tls_cipher derived as RC4-128 IDEA-CBC
                         # ssh_cipher derived as RC4-128 IDEA-CBC
                     ''',
                     TESTSUBPOL1='tls_cipher = NULL+',
                     # BEHAVIOUR CHANGE: same as above.
                     TESTSUBPOL2='cipher = -RC4-128',
                     TESTSUBPOL3='ssh_cipher = +NULL')
    # Used to be ['RC4-128', 'IDEA-CBC', 'NULL']
    assert cp.scoped({'tls', 'openssl'}).enabled['cipher'] == [
        'IDEA-CBC', 'NULL'
    ]
    # Used to be ['NULL', 'RC4-128', 'IDEA-CBC']
    assert cp.scoped({'ssh', 'openssh'}).enabled['cipher'] == [
        'NULL', 'IDEA-CBC'
    ]
    assert cp.scoped().enabled['cipher'] == ['IDEA-CBC']


def test_cryptopolicy_sha1_in_dnssec(tmpdir):
    with pytest.warns(PolicySyntaxDeprecationWarning):
        cp = _policy(tmpdir,
                     TESTPOL='''
                         hash = MD5
                         sha1_in_dnssec = 1
                     ''')
    assert cp.scoped({'tls', 'openssl'}).enabled['hash'] == ['MD5']
    assert cp.scoped({'tls', 'openssl'}).enabled['sign'] == []
    b = cp.scoped({'dnssec', 'bind'})
    assert b.enabled['hash'] == ['MD5', 'SHA1']
    assert b.enabled['sign'] == ['RSA-SHA1', 'ECDSA-SHA1']
    assert 'DSA-SHA1' in b.disabled['sign']


def test_cryptopolicy_prepend_order(tmpdir):
    assert glob('AES-192-*M', 'cipher') == ['AES-192-GCM', 'AES-192-CCM']
    # AES-192-*M expands to AES-192-GCM AES-192-CCM ...
    cp = _policy(tmpdir,
                 TESTPOL='cipher = NULL',
                 # ... but +AES-192-*M expands to +AES-192-CCM +AES-192-GCM
                 # so that -GCM ends up first and has higher priority
                 SUBPOL1='cipher = +AES-192-*M')
    assert cp.scoped({'tls', 'openssl'}).enabled['cipher'] == [
        'AES-192-GCM', 'AES-192-CCM', 'NULL'
    ]


def test_cryptopolicy_no_duplicates(tmpdir):
    cp = _policy(tmpdir, TESTPOL='cipher = AES-192-G* NULL AES-192-C*')
    assert cp.scoped({'tls', 'openssl'}).enabled['cipher'] == [
        'AES-192-GCM', 'NULL', 'AES-192-CCM', 'AES-192-CTR', 'AES-192-CBC'
    ]
    cp = _policy(tmpdir,
                 TESTPOL='cipher = AES-192-G* NULL AES-192-C*',
                 SUBPOL1='cipher = AES-192-CTR+',  # no effect
                 SUBPOL2='cipher = +AES-192-CCM')  # moves CCM first

    assert cp.scoped({'tls', 'openssl'}).enabled['cipher'] == [
        'AES-192-CCM', 'AES-192-GCM', 'NULL', 'AES-192-CTR', 'AES-192-CBC'
    ]


def test_cryptopolicy_minver(tmpdir):
    cp = _policy(tmpdir, TESTPOL='protocol@TLS = TLS*\nmin_tls_version=TLS1.1')
    tls_cp = cp.scoped({'tls', 'openssl'})
    assert tls_cp.enabled['protocol'] == ['TLS1.3', 'TLS1.2', 'TLS1.1']
    assert tls_cp.min_tls_version == 'TLS1.1'
    assert tls_cp.max_tls_version == 'TLS1.3'
    assert tls_cp.min_dtls_version is None
    assert tls_cp.max_dtls_version is None


def test_cryptopolicy_maxver(tmpdir):
    cp = _policy(tmpdir,
                 TESTPOL='protocol@TLS = DTLS*\nmin_dtls_version=DTLS1.0')
    tls_cp = cp.scoped({'tls', 'openssl'})
    assert tls_cp.enabled['protocol'] == ['DTLS1.2', 'DTLS1.0']
    assert tls_cp.min_dtls_version == 'DTLS1.0'
    assert tls_cp.max_dtls_version == 'DTLS1.2'
    assert tls_cp.min_tls_version is None
    assert tls_cp.max_tls_version is None


def test_cryptopolicy_to_string_empty(tmpdir):
    reference = textwrap.dedent('''
        # Policy EMPTYPOL:EMPTYSUBPOL1:EMPTYSUBPOL2 dump
        #
        # Do not parse the contents of this file with automated tools,
        # it is provided for review convenience only.
        #
        # Baseline values for all scopes:
        cipher =
        group =
        hash =
        key_exchange =
        mac =
        protocol =
        sign =
        arbitrary_dh_groups = 0
        min_dh_size = 0
        min_dsa_size = 0
        min_rsa_size = 0
        sha1_in_certs = 0
        ssh_certs = 0
        ssh_etm = 0
        # No scope-specific properties found.
    ''').lstrip()
    cp = _policy(tmpdir, EMPTYPOL='', EMPTYSUBPOL1='\n', EMPTYSUBPOL2='\t')
    print(repr(reference[:30]))
    print(repr(str(cp)[:30]))
    assert str(cp) == reference


def test_cryptopolicy_to_string_twisted(tmpdir):
    reference = textwrap.dedent('''
        # Policy TESTPOL dump
        #
        # Do not parse the contents of this file with automated tools,
        # it is provided for review convenience only.
        #
        # Baseline values for all scopes:
        cipher = RC4-128 IDEA-CBC
        group =
        hash = MD5
        key_exchange =
        mac =
        protocol =
        sign =
        arbitrary_dh_groups = 0
        min_dh_size = 0
        min_dsa_size = 0
        min_rsa_size = 0
        sha1_in_certs = 0
        ssh_certs = 0
        ssh_etm = 0
        # Scope-specific properties derived for select backends:
        cipher@gnutls = DES-CBC
        hash@gnutls =
        sha1_in_certs@gnutls = 1
        cipher@java-tls = DES-CBC
        cipher@nss = DES-CBC
        cipher@openssl = NULL DES-CBC
    ''').lstrip()
    cp = _policy(tmpdir,
                 TESTPOL='''
                     hash = MD5
                     cipher@openssl = SEED-CBC  # overridden in the next line
                     cipher = RC4-128 IDEA-CBC
                     cipher@tls = DES-CBC
                     cipher@openssl = +NULL
                     sha1_in_certs@gnutls = 1
                     hash@gnutls = -MD5
                 ''')
    assert str(cp) == reference



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2019 Red Hat, Inc.
# Copyright (c) 2019 Tomáš Mráz <tmraz@fedoraproject.org>

from .cryptopolicies import UnscopedCryptoPolicy

__all__ = ['UnscopedCryptoPolicy']



#!/usr/bin/python3

import sys

# krb5check can be found at:
#     https://github.com/frozencemetery/krb5check
# At a minimum, it verifies that:
#   - no unrecognized sections are specified
#   - no values are overriding each other
#   - libdefaults section is present
#   - permitted_enctypes is specified
#   - no unknown enctypes are specified
#   - no known-broken enctypes are specified
#   - if pkinit_dh_min_bits is specified, it is larger than default
#   - if pkinit_dh_min_bits is specified, it is reasonable

try:
    from krb5check.krb5_conf import parse, check, ACCEPTED_ENCTYPES
except ImportError:
    print("Skipping krb5 test; checker not found!")
    sys.exit(0)

print("Checking the Krb5 configuration")

# Don't verify EMPTY policy
for policy in ["LEGACY", "DEFAULT", "FUTURE", "FIPS"]:
    path = 'tests/outputs/' + policy + '-krb5.txt'

    sections = parse(path)
    check(sections, ACCEPTED_ENCTYPES)



#!/usr/bin/python3

import ctypes
import ctypes.util
import glob
import os
import shutil
import subprocess
import sys


if shutil.which('nss-policy-check') is None:
    print('nss-policy-check not found, skipping check', file=sys.stderr)
    sys.exit(0)


# Cannot validate with pre-3.59 NSS that doesn't know ECDSA/RSA-PSS/RSA-PKCS
# identifiers yet. Checking for 3.66 because that's what we'll have in 8.5.
try:
    nss = ctypes.CDLL(ctypes.util.find_library('nss3'))
    if not nss.NSS_VersionCheck(b'3.66'):
        print('Skipping nss-policy-check verification '
              'due to nss being older than 3.66', file=sys.stderr)
        sys.exit(0)
except AttributeError:
    print('Cannot determine nss version with ctypes, hoping for >=3.59',
          file=sys.stderr)


print('Checking the NSS configuration')

for policy_path in glob.glob('tests/outputs/*-nss.txt'):
    policy = os.path.basename(policy_path)[:-len('-nss.txt')]
    print(f'Checking policy {policy}')
    if policy not in ('EMPTY', 'GOST-ONLY'):
        p = subprocess.Popen(['nss-policy-check', policy_path],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
        output, _ = p.communicate()
        if p.wait():
            print(f'Error in NSS policy for {policy}')
            print(f'NSS policy for {policy}:', file=sys.stderr)
            with open(policy_path) as policy_file:
                shutil.copyfileobj(policy_file, sys.stderr)
                sys.stderr.write('\n')
            print('nss-policy-check error:', file=sys.stderr)
            print(output.decode(), file=sys.stderr)
            sys.exit(1)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2019 Red Hat, Inc.
# Copyright (c) 2019 Tomáš Mráz <tmraz@fedoraproject.org>

from .configgenerator import ConfigGenerator


class KRB5Generator(ConfigGenerator):
	CONFIG_NAME = 'krb5'
	SCOPES = {'kerberos', 'krb5'}

	cipher_map = {
		'AES-256-CTR':'',
		'AES-128-CTR':'',
		'AES-256-GCM':'',
		'AES-256-CCM':'',
		'CHACHA20-POLY1305':'',
		'CAMELLIA-256-GCM':'',
		'AES-128-GCM':'',
		'AES-128-CCM':'',
		'CAMELLIA-128-GCM':'',

		'AES-256-CBC':'aes256-cts-hmac-sha1-96 aes256-cts-hmac-sha384-192',
		'AES-128-CBC':'aes128-cts-hmac-sha1-96 aes128-cts-hmac-sha256-128',
		'CAMELLIA-256-CBC':'camellia256-cts-cmac',
		'CAMELLIA-128-CBC':'camellia128-cts-cmac',
		'RC4-128':'arcfour-hmac-md5',
		'DES-CBC':'',
		'CAMELLIA-128-CTS':'camellia128-cts-cmac',
		'3DES-CBC':''
	}

	@classmethod
	def generate_config(cls, policy):
		p = policy.enabled
		sep = ' '

		cfg = '[libdefaults]\n'
		cfg += 'permitted_enctypes = '
		s = ''
		for i in p['cipher']:
			try:
				s = cls.append(s, cls.cipher_map[i], sep)
			except KeyError:
				pass
		if 'RC4-128' in p['cipher'] and 'MD5' in p['hash']:
			s = cls.append(s, 'arcfour-hmac-md5', sep)

		cfg += s + '\n'

		# By default libkrb5 sets the min_bits to 2048, don't
		# go lower than that.
		if policy.integers['min_dh_size'] > 2048:
			# $string .= "pkinit_dh_min_bits=$min_dh_size\n";
			# krb5.conf only accepts 2048 or 4096
			cfg += 'pkinit_dh_min_bits=4096\n'

		return cfg

	@classmethod
	def test_config(cls, config):  # pylint: disable=unused-argument
		return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

from . import alg_lists, rules, scope
from .general import PolicySyntaxError, PolicyFileNotFoundError

__all__ = [
    'alg_lists', 'rules', 'scope',
    'PolicySyntaxError', 'PolicyFileNotFoundError'
]



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

class PolicySyntaxError(ValueError, UserWarning):
    pass


class PolicyFileNotFoundError(FileNotFoundError):
    def __init__(self, pname, fname, paths):
        super().__init__(f'Unknown policy `{pname}`: '
                         f'file `{fname}` not found in ({", ".join(paths)})')



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2019 Red Hat, Inc.
# Copyright (c) 2019 Tomáš Mráz <tmraz@fedoraproject.org>

import collections
import enum
import fnmatch
import os
import re
import warnings

from . import alg_lists
from . import validation  # moved out of the way to not obscure the flow


# Defaults of integer property values (doubles as an allowlist)

INT_DEFAULTS = {k: 0 for k in (
    'arbitrary_dh_groups',
    'min_dh_size', 'min_dsa_size', 'min_rsa_size',
    'sha1_in_certs',
    'ssh_certs', 'ssh_etm',
)}


# Scopes (`@!ipsec`) and matching them

SCOPE_ANY = '*'

ALL_SCOPES = (  # defined explicitly to catch typos / globbing nothing
    'tls', 'ssl', 'openssl', 'nss', 'gnutls', 'java-tls',
    'ssh', 'openssh', 'openssh-server', 'openssh-client', 'libssh',
    'ipsec', 'ike', 'libreswan',
    'kerberos', 'krb5',
    'dnssec', 'bind',
)
DUMPABLE_SCOPES = {  # TODO: fix duplication, backends specify same things
    'bind': {'bind', 'dnssec'},
    'gnutls': {'gnutls', 'tls', 'ssl'},
    'java-tls': {'java-tls', 'tls', 'ssl'},
    'krb5': {'krb5', 'kerberos'},
    'libreswan': {'ipsec', 'ike', 'libreswan'},
    'libssh': {'libssh', 'openssh', 'ssh'},
    'nss': {'nss', 'tls', 'ssl'},
    'openssh-client': {'openssh-client', 'openssh', 'ssh'},
    'openssh-server': {'openssh-server', 'openssh', 'ssh'},
    'openssl': {'openssl', 'tls', 'ssl'},
}


class ScopeSelector:
    def __init__(self, pattern=SCOPE_ANY):
        """
        Initialize a scope selector.
        An example would be `ssh` in `ciphers@ssh = -NULL`.
        When openssh backend will request the configuration,
        it'll offer (`{'ssh', 'openssh'}`) as scopes
        and the rule above will be taken into account.
        Both patterns and scopes are cast to lowercase.
        For more examples, refer to tests/unit/parsing/test_scope_selector.py
        >>> ss = ScopeSelector('!{SSH,IPsec}')
        >>> ss.matches({'ipsec', 'libreswan'})
        False
        >>> ss.matches({'tls', 'openssl'})
        True
        """
        self.pattern = pattern = pattern.lower()
        self._positive = not pattern.startswith('!')
        p = pattern if self._positive else pattern[1:]

        validation.scope.illegal_characters(p, original_pattern=self.pattern)
        validation.scope.curly_brackets(p, original_pattern=self.pattern)

        self._globs = p[1:-1].split(',') if p.startswith('{') else [p]

        validation.scope.resulting_globs(self._globs, ALL_SCOPES,
                                         original_pattern=self.pattern)

    def __str__(self):
        return f'<ScopeSelector pattern={repr(self.pattern)}>'

    def matches(self, scopes):
        """
        Checks whether ScopeSelector matches one of the scopes.
        For more examples, refer to tests/unit/parsing/test_scope_selector.py
        >>> ScopeSelector('{SSH,IPsec}').matches({'ipsec', 'libreswan'})
        True
        >>> ScopeSelector('!{SSH,IPsec}').matches({'ipsec', 'libreswan'})
        False
        """
        if self.pattern == SCOPE_ANY:  # matches even an empty set
            return True
        scopes = [s.lower() for s in scopes]
        assert all(s in ALL_SCOPES for s in scopes)  # supplied by backends
        if self._positive:
            return any(fnmatch.filter(scopes, g) for g in self._globs)
        return all(not fnmatch.filter(scopes, g) for g in self._globs)


# Operations: interpreting right hand sides of (sub)policy files

class Operation(enum.Enum):
    """
    An operation that comes with the right-hand value of the directive.
    """
    RESET = 1     # cipher =
    PREPEND = 2   # cipher = +NULL
    APPEND = 3    # cipher = NULL+
    OMIT = 4      # cipher = -NULL
    SET_INT = 5   # sha1_in_certs = 0; setting to something that's all digits


def parse_rhs(rhs, prop_name):
    """
    Parses right-hand parts of the directives
    into lists of operation/value pairs.
    For more examples, refer to tests/unit/test_parsing.py
    >>> parse_rhs('', 'cipher')
    [(<Operation.RESET: 1>, None)]
    >>> parse_rhs('IDEA-CBC SEED-CBC', 'cipher')
    [(<Operation.RESET: 1>, None),
     (<Operation.APPEND: 3>, 'IDEA-CBC'),
     (<Operation.APPEND: 3>, 'SEED-CBC')]
    >>> # 3DES-CBC gets prepended last for higher prio
    >>> parse_rhs('+*DES-CBC', 'cipher')
    [(<Operation.PREPEND: 2>, 'DES-CBC'),
     (<Operation.PREPEND: 2>, '3DES-CBC')]
    """
    def differential(v):
        return v.startswith('+') or v.endswith('+') or v.startswith('-')

    if rhs.isdigit():
        if prop_name not in alg_lists.ALL and prop_name in INT_DEFAULTS:
            return [(Operation.SET_INT, int(rhs))]
        elif prop_name in alg_lists.ALL:
            raise validation.rules.NonIntPropertyIntValueError(prop_name)
        else:
            assert prop_name not in alg_lists.ALL
            assert prop_name not in INT_DEFAULTS
            # pass for now, it's gonna be caught as non-existing algclass
    else:
        if prop_name in INT_DEFAULTS:
            raise validation.rules.IntPropertyNonIntValueError(prop_name)

    values = rhs.split()

    if not any(differential(v) for v in values):  # Setting something anew
        values = sum([alg_lists.glob(v, prop_name) for v in values], [])
        return ([(Operation.RESET, None)]
                + [(Operation.APPEND, v) for v in values])
    elif all(differential(v) for v in values):  # Modifying an existing list
        operations = []
        for value in values:
            if value.startswith('+'):
                op = Operation.PREPEND
                unglob = alg_lists.glob(value[1:], prop_name)[::-1]
            elif value.endswith('+'):
                op = Operation.APPEND
                unglob = alg_lists.glob(value[:-1], prop_name)[::-1]
            else:
                assert value.startswith('-')
                op = Operation.OMIT
                unglob = alg_lists.glob(value[1:], prop_name)
            operations.extend([(op, v) for v in unglob])
        return operations
    else:  # Forbidden to mix them on one line
        raise validation.rules.MixedDifferentialNonDifferentialError(rhs)


# Directives: interpreting lines of (sub)policy files

Directive = collections.namedtuple('Directive', (
    'prop_name', 'scope', 'operation', 'value'
))


def parse_line(line):
    """
    Parses configuration lines into tuples of directives.
    For more examples, refer to tests/unit/test_parsing.py
    >>> parse_line('cipher@TLS = RC4* NULL')
    [Directive(prop_name='cipher', scope='tls',
               operation=<Operation.RESET: 1>, value=None),
     Directive(prop_name='cipher', scope='tls',
               operation=<Operation.APPEND: 3>, value='RC4-40'),
     Directive(prop_name='cipher', scope='tls',
               operation=<Operation.APPEND: 3>, value='RC4-128'),
     Directive(prop_name='cipher', scope='tls',
               operation=<Operation.APPEND: 3>, value='NULL')]
    """
    if not line.strip():
        return []
    validation.rules.count_equals_signs(line)

    lhs, rhs = line.split('=')
    lhs, rhs = lhs.strip(), rhs.strip()
    validation.rules.empty_lhs(lhs, line)

    prop_name, scope = lhs.split('@', 1) if '@' in lhs else (lhs, SCOPE_ANY)

    return [Directive(prop_name=prop_name, scope=scope.lower(),
                      operation=operation, value=value)
            for operation, value in parse_rhs(rhs, prop_name)]


def syntax_check_line(line, warn=False):
    try:
        l = parse_line(line)
        for d in l:
            ScopeSelector(d.scope)  # attempt parsing
    except validation.PolicySyntaxError as ex:
        if not warn:
            raise
        warnings.warn(ex)


class PolicySyntaxDeprecationWarning(FutureWarning):
    def __init__(self, deprecated, replacement):
        replacement = replacement.replace('\n', ' and ')
        msg = f'option {deprecated} is deprecated'
        msg += f', please rewrite your rules using {replacement}; '
        msg += 'be advised that it is not always a 1-1 replacement'
        super().__init__(msg)


def preprocess_text(text):
    r"""
    Preprocesses text before parsing.
    Fixes line breaks, handles backwards compatibility.
    >>> preprocess_text('cipher = c1 \\ \nc2#x')
    'cipher = c1 c2'
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     preprocess_text('ike_protocol = IKEv2')
    'protocol@IKE = IKEv2'
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     preprocess_text('min_tls_version=TLS1.3')
    'protocol@TLS = -SSL2.0 -SSL3.0 -TLS1.0 -TLS1.1 -TLS1.2'
    """
    text = re.sub(r'#.*', '', text)
    text = text.replace('=', ' = ')
    text = '\n'.join((l.strip() for l in text.split('\n')))
    text = text.replace('\\\n', '')
    text = '\n'.join((l.strip() for l in text.split('\n')))
    text = '\n'.join((re.sub(r'\s+', ' ', l) for l in text.split('\n')))
    text = re.sub('\n+', '\n', text).strip()

    if re.findall(r'\bprotocol\s*=', text):
        warnings.warn(PolicySyntaxDeprecationWarning('protocol',
                                                     'protocol@TLS'))

    POSTFIX_REPLACEMENTS = {
        'tls_cipher': 'cipher@TLS',
        'ssh_cipher': 'cipher@SSH',
        'ssh_group': 'group@SSH',
        'ike_protocol': 'protocol@IKE',
    }
    for fr, to in POSTFIX_REPLACEMENTS.items():
        regex = r'\b' + fr + r'\s*=(.*)'
        ms = re.findall(regex, text)
        if ms:
            warnings.warn(PolicySyntaxDeprecationWarning(fr, to))
        text = re.sub(regex, '', text)
        for m in ms:
            text += f'\n\n{to} ={m}'
    text = re.sub('\n+', '\n', text).strip()

    PLAIN_REPLACEMENTS = {
        'sha1_in_dnssec = 0':
            'hash@DNSSec = -SHA1\nsign@DNSSec = -RSA-SHA1 -ECDSA-SHA1',
        'sha1_in_dnssec = 1':
            'hash@DNSSec = SHA1+\nsign@DNSSec = RSA-SHA1+ ECDSA-SHA1+',
    }
    for fr, to in PLAIN_REPLACEMENTS.items():
        regex = r'\b' + fr + r'\b'
        if re.search(regex, text):
            warnings.warn(PolicySyntaxDeprecationWarning(fr, to))
        text = re.sub(regex, to, text)

    dtls_versions = list(alg_lists.DTLS_PROTOCOLS[::-1])
    while dtls_versions:
        neg = " ".join(("-" + v for v in dtls_versions[:-1]))
        text = re.sub(r'\bmin_dtls_version = ' + dtls_versions[-1] + r'\b',
                      f'protocol@TLS = {neg}' if neg else '', text)
        dtls_versions.pop()
    text = re.sub(r'\bmin_dtls_version = 0\b', '', text)

    tls_versions = list(alg_lists.TLS_PROTOCOLS[::-1])
    while tls_versions:
        neg = " ".join(("-" + v for v in tls_versions[:-1]))
        text = re.sub(r'\bmin_tls_version = ' + tls_versions[-1] + r'\b',
                      f'protocol@TLS = {neg}' if neg else '', text)
        tls_versions.pop()
    text = re.sub(r'\bmin_tls_version = 0\b', '', text)

    return text


# Finally, constructing a policy

class ScopedPolicy:
    """
    An entity constructing lists of what's `.enabled` and what's `.disabled`
    when the given scopes are active.
    >>> sp = ScopedPolicy(parse_line('cipher@TLS = RC4* NULL'), {'tls'})
    >>> 'AES-256-GCM' in sp.disabled['cipher']
    True
    >>> sp.enabled['cipher']
    ['RC4-40', 'RC4-128', 'NULL']
    >>> ScopedPolicy(parse_line('min_dh_size=2048')).integers['min_dh_size']
    2048
    """
    def __init__(self, directives, relevant_scopes=None):
        relevant_scopes = relevant_scopes or set()
        self.integers = INT_DEFAULTS.copy()
        self.enabled = {prop_name: [] for prop_name in alg_lists.ALL}

        for directive in directives:
            # TODO: validate that the target exists
            ss = ScopeSelector(directive.scope)
            if ss.matches(relevant_scopes):
                if directive.operation == Operation.RESET:
                    self.enabled[directive.prop_name] = []
                elif directive.operation == Operation.APPEND:
                    enabled = self.enabled[directive.prop_name]
                    if directive.value not in enabled:
                        enabled.append(directive.value)
                elif directive.operation == Operation.PREPEND:
                    enabled = self.enabled[directive.prop_name]
                    # in case of duplicates, remove the latter, lower-prio ones
                    if directive.value in enabled:
                        enabled.remove(directive.value)
                    enabled.insert(0, directive.value)
                elif directive.operation == Operation.OMIT:
                    self.enabled[directive.prop_name] = [
                        e for e in self.enabled[directive.prop_name]
                        if e != directive.value
                    ]
                else:
                    assert directive.operation == Operation.SET_INT
                    self.integers[directive.prop_name] = directive.value
        assert len(self.enabled) == len(set(self.enabled))

        self.disabled = {prop_name: [e for e in alg_lists.ALL[prop_name]
                                     if e not in self.enabled[prop_name]]
                         for prop_name in alg_lists.ALL}

    @property
    def min_tls_version(self):
        return alg_lists.min_tls_version(self.enabled['protocol'])

    @property
    def max_tls_version(self):
        return alg_lists.max_tls_version(self.enabled['protocol'])

    @property
    def min_dtls_version(self):
        return alg_lists.min_dtls_version(self.enabled['protocol'])

    @property
    def max_dtls_version(self):
        return alg_lists.max_dtls_version(self.enabled['protocol'])


# Locating policy files

def lookup_file(policyname, fname, paths):
    for d in paths:
        p = os.path.join(d, fname)
        if os.access(p, os.R_OK):
            return p
    raise validation.PolicyFileNotFoundError(policyname, fname, paths)


# main class

class UnscopedCryptoPolicy:
    CONFIG_DIR = '/etc/crypto-policies'

    SHARE_DIR = '/usr/share/crypto-policies'

    def __init__(self, policy_name, *subpolicy_names, policydir=None):
        self.policydir = policydir
        self.policyname = ':'.join((policy_name,) + subpolicy_names)

        self.lines = []

        directives = self.read_policy_file(policy_name)
        for subpolicy_name in subpolicy_names:
            directives += self.read_policy_file(subpolicy_name, subpolicy=True)
        self._directives = directives

    def is_empty(self):
        return not self._directives

    def scoped(self, scopes=None):
        return ScopedPolicy(self._directives, scopes or {})

    def read_policy_file(self, name, subpolicy=False):
        pdir = self.policydir or 'policies'
        if subpolicy:
            pdir = os.path.join(pdir, 'modules')
        p = lookup_file(name,
                        name + ('.pol' if not subpolicy else '.pmod'), (
                            os.path.curdir,
                            pdir,
                            os.path.join(self.CONFIG_DIR, pdir),
                            os.path.join(self.SHARE_DIR, pdir),
                        ))
        # TODO: error handling
        with open(p) as f:
            text = f.read()
        text = preprocess_text(text)
        lines = text.split('\n')
        for l in lines:  # display several warnings at once
            syntax_check_line(l, warn=True)
        for l in lines:  # crash
            syntax_check_line(l)
        return sum([parse_line(l) for l in lines], [])

    def __str__(self):
        def fmt(key, value):
            s = ' '.join(value) if isinstance(value, list) else str(value)
            return f'{key} = {s}'.rstrip() + '\n'

        generic_scoped = self.scoped()
        s = f'# Policy {self.policyname} dump\n'
        s += '#\n'
        s += '# Do not parse the contents of this file with automated tools,\n'
        s += '# it is provided for review convenience only.\n'
        s += '#\n'
        s += '# Baseline values for all scopes:\n'
        generic_all = {**generic_scoped.enabled, **generic_scoped.integers}
        for prop_name, value in generic_all.items():
            s += fmt(prop_name, value)
        anything_scope_specific = False
        for scope_name, scope_set in DUMPABLE_SCOPES.items():
            specific_scoped = self.scoped(scopes=scope_set)
            specific_all = {**specific_scoped.enabled,
                            **specific_scoped.integers}
            for prop_name, value in specific_all.items():
                if value != generic_all[prop_name]:
                    if not anything_scope_specific:
                        s += ('# Scope-specific properties '
                              'derived for select backends:\n')
                        anything_scope_specific = True
                    s += fmt(f'{prop_name}@{scope_name}', value)
        if not anything_scope_specific:
            s += '# No scope-specific properties found.\n'
        return s



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2019 Red Hat, Inc.
# Copyright (c) 2019 Tomáš Mráz <tmraz@fedoraproject.org>

from subprocess import call, CalledProcessError
from tempfile import mkstemp

import os

from .configgenerator import ConfigGenerator


class OpenSSHGenerator(ConfigGenerator):
	_FORMAT_STRING = ''

	cipher_map = {
		'AES-256-GCM':'aes256-gcm@openssh.com',
		'AES-256-CTR':'aes256-ctr',
		'AES-192-GCM':'',  # not supported
		'AES-192-CTR':'aes192-ctr',
		'AES-128-GCM':'aes128-gcm@openssh.com',
		'AES-128-CTR':'aes128-ctr',
		'CHACHA20-POLY1305':'chacha20-poly1305@openssh.com',
		'CAMELLIA-256-GCM':'',
		'AES-256-CCM':'',
		'AES-192-CCM':'',
		'AES-128-CCM':'',
		'CAMELLIA-128-GCM':'',
		'AES-256-CBC':'aes256-cbc',
		'AES-192-CBC':'aes192-cbc',
		'AES-128-CBC':'aes128-cbc',
		'CAMELLIA-256-CBC':'',
		'CAMELLIA-128-CBC':'',
		'RC4-128':'',
		'DES-CBC':'',
		'CAMELLIA-128-CTS':'',
		'3DES-CBC':'3des-cbc'
	}

	mac_map_etm = {
		'HMAC-MD5':'hmac-md5-etm@openssh.com',
		'UMAC-64':'umac-64-etm@openssh.com',
		'UMAC-128':'umac-128-etm@openssh.com',
		'HMAC-SHA1':'hmac-sha1-etm@openssh.com',
		'HMAC-SHA2-256':'hmac-sha2-256-etm@openssh.com',
		'HMAC-SHA2-512':'hmac-sha2-512-etm@openssh.com'
	}

	mac_map = {
		'HMAC-MD5':'hmac-md5',
		'UMAC-64':'umac-64@openssh.com',
		'UMAC-128':'umac-128@openssh.com',
		'HMAC-SHA1':'hmac-sha1',
		'HMAC-SHA2-256':'hmac-sha2-256',
		'HMAC-SHA2-512':'hmac-sha2-512'
	}

	kx_map = {
		'ECDHE-SECP521R1-SHA2-512':'ecdh-sha2-nistp521',
		'ECDHE-SECP384R1-SHA2-384':'ecdh-sha2-nistp384',
		'ECDHE-SECP256R1-SHA2-256':'ecdh-sha2-nistp256',
		'ECDHE-X25519-SHA2-256':'curve25519-sha256,curve25519-sha256@libssh.org',
		'DHE-FFDHE-1024-SHA1':'diffie-hellman-group1-sha1',
		'DHE-FFDHE-2048-SHA1':'diffie-hellman-group14-sha1',
		'DHE-FFDHE-2048-SHA2-256':'diffie-hellman-group14-sha256',
		'DHE-FFDHE-4096-SHA2-512':'diffie-hellman-group16-sha512',
		'DHE-FFDHE-8192-SHA2-512':'diffie-hellman-group18-sha512',
	}

	gx_map = {
		'DHE-SHA1':'diffie-hellman-group-exchange-sha1',
		'DHE-SHA2-256':'diffie-hellman-group-exchange-sha256',
	}

	gss_kx_map = {
		'DHE-GSS-SHA1':'gss-gex-sha1-',
		'DHE-GSS-FFDHE-1024-SHA1':'gss-group1-sha1-',
		'DHE-GSS-FFDHE-2048-SHA1':'gss-group14-sha1-',
		'DHE-GSS-FFDHE-2048-SHA2-256':'gss-group14-sha256-',
		'ECDHE-GSS-SECP256R1-SHA2-256':'gss-nistp256-sha256-',
		'ECDHE-GSS-X25519-SHA2-256':'gss-curve25519-sha256-',
		'DHE-GSS-FFDHE-4096-SHA2-512':'gss-group16-sha512-',
	}

	sign_map = {
		'RSA-SHA1':'ssh-rsa',
		'DSA-SHA1':'ssh-dss',
		'RSA-SHA2-256':'rsa-sha2-256',
		'RSA-SHA2-512':'rsa-sha2-512',
		'ECDSA-SHA2-256':'ecdsa-sha2-nistp256',
		'ECDSA-SHA2-384':'ecdsa-sha2-nistp384',
		'ECDSA-SHA2-512':'ecdsa-sha2-nistp521',
		'EDDSA-ED25519':'ssh-ed25519',
	}

	sign_map_certs = {
		'RSA-SHA1':'ssh-rsa-cert-v01@openssh.com',
		'DSA-SHA1':'ssh-dss-cert-v01@openssh.com',
		'RSA-SHA2-256':'rsa-sha2-256-cert-v01@openssh.com',
		'RSA-SHA2-512':'rsa-sha2-512-cert-v01@openssh.com',
		'ECDSA-SHA2-256':'ecdsa-sha2-nistp256-cert-v01@openssh.com',
		'ECDSA-SHA2-384':'ecdsa-sha2-nistp384-cert-v01@openssh.com',
		'ECDSA-SHA2-512':'ecdsa-sha2-nistp521-cert-v01@openssh.com',
		'EDDSA-ED25519':'ssh-ed25519-cert-v01@openssh.com',
	}

	@classmethod
	def generate_options(cls, policy, local_kx_map, local_gss_kx_map, do_host_key):
		p = policy.enabled
		cfg = ''
		sep = ','

		s = ''
		for i in p['cipher']:
			try:
				s = cls.append(s, cls.cipher_map[i], sep)
			except KeyError:
				pass

		if s:
			cfg += cls._FORMAT_STRING.format('Ciphers', s)

		s = ''
		if policy.integers['ssh_etm']:
			for i in p['mac']:
				try:
					s = cls.append(s, cls.mac_map_etm[i], sep)
				except KeyError:
					pass
		for i in p['mac']:
			try:
				s = cls.append(s, cls.mac_map[i], sep)
			except KeyError:
				pass

		if s:
			cfg += cls._FORMAT_STRING.format('MACs', s)

		s = ''
		gss = ''
		for kx in p['key_exchange']:
			for h in p['hash']:
				if policy.integers['arbitrary_dh_groups']:
					try:
						val = cls.gx_map[kx + '-' + h]
						s = cls.append(s, val, sep)
					except KeyError:
						pass
					try:
						val = local_gss_kx_map[kx + '-' + h]
						gss = cls.append(gss, val, sep)
					except KeyError:
						pass
				for g in p['group']:
					try:
						val = local_kx_map[kx + '-' + g + '-' + h]
						s = cls.append(s, val, sep)
					except KeyError:
						pass
					try:
						val = local_gss_kx_map[kx + '-' + g + '-' + h]
						gss = cls.append(gss, val, sep)
					except KeyError:
						pass

		if gss:
			cfg += cls._FORMAT_STRING.format('GSSAPIKexAlgorithms', gss)
		else:
			cfg += cls._FORMAT_STRING.format('GSSAPIKeyExchange', 'no')

		if s:
			cfg += cls._FORMAT_STRING.format('KexAlgorithms', s)

		s = ''
		for i in p['sign']:
			try:
				s = cls.append(s, cls.sign_map[i], sep)
			except KeyError:
				pass
			if policy.integers['ssh_certs'] == 1:
				try:
					s = cls.append(s, cls.sign_map_certs[i], sep)
				except KeyError:
					pass

		if s:
			# As OpenSSH currently ignores existing known host
			# entries with this setting we cannot use it on client.
			# Otherwise we could break existing users.
			if do_host_key:
				cfg += cls._FORMAT_STRING.format('HostKeyAlgorithms', s)
			cfg += cls._FORMAT_STRING.format('PubkeyAcceptedKeyTypes', s)

		s = ''
		for i in p['sign']:
			try:
				s = cls.append(s, cls.sign_map[i], sep)
			except KeyError:
				pass

		if s:
			cfg += cls._FORMAT_STRING.format('CASignatureAlgorithms', s)

		return cfg


class OpenSSHClientGenerator(OpenSSHGenerator):
	CONFIG_NAME = 'openssh'
	SCOPES = {'ssh', 'openssh', 'openssh-client'}

	_FORMAT_STRING = '{0} {1}\n'

	@classmethod
	def generate_config(cls, policy):
		local_kx_map = dict(cls.kx_map)
		local_gss_kx_map = dict(cls.gss_kx_map)

		return cls.generate_options(policy, local_kx_map, local_gss_kx_map, False)

	@classmethod
	def test_config(cls, config):
		if not os.access('/usr/bin/ssh', os.X_OK):
			return True

		fd, path = mkstemp()

		ret = 255
		try:
			with os.fdopen(fd, 'w') as f:
				f.write(config)
			try:
				ret = call('/usr/bin/ssh -G -F ' + path +
					' bogus654_server >/dev/null',
					shell=True)
			except CalledProcessError:
				cls.eprint("/usr/bin/ssh: Execution failed")
		finally:
			os.unlink(path)

		if ret:
			cls.eprint("There is an error in OpenSSH generated policy")
			cls.eprint("Policy:\n%s" % config)
			return False
		return True


class OpenSSHServerGenerator(OpenSSHGenerator):
	CONFIG_NAME = 'opensshserver'
	SCOPES = {'ssh', 'openssh', 'openssh-server'}

	# We need restart here, since systemd needs to pick up new command line options
	RELOAD_CMD = 'systemctl try-restart sshd.service 2>/dev/null || :\n'

	_FORMAT_STRING = '-o{0}={1} '

	@classmethod
	def generate_config(cls, policy):
		# Difference from client, keep group1 disabled on server
		local_kx_map = dict(cls.kx_map)
		local_gss_kx_map = dict(cls.gss_kx_map)
		del local_kx_map['DHE-FFDHE-1024-SHA1']
		del local_gss_kx_map['DHE-GSS-FFDHE-1024-SHA1']

		cfg = cls.generate_options(policy, local_kx_map, local_gss_kx_map, True)
		cfg = cfg.rstrip()
		return 'CRYPTO_POLICY=\'' + cfg + '\''

	@classmethod
	def _test_setup(cls):
		_fd, path = mkstemp()
		os.unlink(path)
		ret = 255
		try:
			ret = call('/usr/bin/ssh-keygen -t rsa -b 2048 -f ' + path +
				' -N "" >/dev/null',
				shell=True)
		except CalledProcessError:
			cls.eprint("/usr/bin/ssh-keygen: Execution failed")

		if ret:
			cls.eprint("SSH Keygen failed when testing OpenSSH server policy")
			return ''
		return path

	@classmethod
	def _test_cleanup(cls, path):
		if path:
			os.unlink(path)

	@classmethod
	def test_config(cls, config):
		if not os.access('/usr/sbin/sshd', os.X_OK):
			return True

		host_key_filename = cls._test_setup()
		if not host_key_filename:
			return False

		fd, path = mkstemp()

		ret = 255
		try:
			with os.fdopen(fd, 'w') as f:
				f.write(config)
			try:
				ret = call('/usr/bin/bash -c \'source ' + path +
					' && /usr/sbin/sshd -T $CRYPTO_POLICY -h ' +
					host_key_filename +
					' -f /dev/null\' >/dev/null',
					shell=True)
			except CalledProcessError:
				cls.eprint("/usr/sbin/sshd: Execution failed")
		finally:
			os.unlink(path)
			cls._test_cleanup(host_key_filename)

		if ret:
			cls.eprint("There is an error in OpenSSH server generated policy")
			cls.eprint("Policy:\n%s" % config)
			return False
		return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

import pytest

from python.cryptopolicies.cryptopolicies import (
    Operation, Directive, parse_line
)

from python.cryptopolicies.validation.rules import MalformedLine


def test_parse_line():
    assert parse_line('cipher = AES-128-GCM AES-256-GCM') == [
        Directive(prop_name='cipher', scope='*',
                  operation=Operation.RESET, value=None),
        Directive(prop_name='cipher', scope='*',
                  operation=Operation.APPEND, value='AES-128-GCM'),
        Directive(prop_name='cipher', scope='*',
                  operation=Operation.APPEND, value='AES-256-GCM'),
    ]
    assert parse_line('cipher@gnutls = +AES-128-GCM') == [
        Directive(prop_name='cipher', scope='gnutls',
                  operation=Operation.PREPEND, value='AES-128-GCM'),
    ]
    assert parse_line('cipher@*SSH = AES-128-CBC+ -NULL') == [
        Directive(prop_name='cipher', scope='*ssh',
                  operation=Operation.APPEND, value='AES-128-CBC'),
        Directive(prop_name='cipher', scope='*ssh',
                  operation=Operation.OMIT, value='NULL'),
    ]
    assert parse_line('cipher =') == [
        Directive(prop_name='cipher', scope='*',
                  operation=Operation.RESET, value=None),
    ]
    assert parse_line('\t\t') == []


def test_parse_bad():
    with pytest.raises(MalformedLine):
        parse_line('a = b = c')
    with pytest.raises(MalformedLine):
        parse_line('test')
    with pytest.raises(MalformedLine):
        parse_line('=4')
    with pytest.raises(MalformedLine):
        parse_line('=')



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2019 Red Hat, Inc.
# Copyright (c) 2019 Tomáš Mráz <tmraz@fedoraproject.org>

import sys


class ConfigGenerator:
	RELOAD_CMD = ''

	@staticmethod
	def append(s, val, sep=':'):
		if s:
			if val:
				return s + sep + val

			return s

		return val

	@staticmethod
	def eprint(*args, **kwargs):
		print(*args, file=sys.stderr, **kwargs)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

from .general import PolicySyntaxError


class AlgorithmClassSyntaxError(PolicySyntaxError):
    pass


class AlgorithmClassUnknownError(AlgorithmClassSyntaxError):
    def __init__(self, alg_class):
        # The wording follows the previous versions
        super().__init__(f'Unknown policy property: `{alg_class}`')


class AlgorithmEmptyMatchError(AlgorithmClassSyntaxError):
    def __init__(self, glob, alg_class):
        # The wording follows the previous versions
        super().__init__(f'Bad value of policy property `{alg_class}`: '
                         f'`{glob}`')



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2019 Red Hat, Inc.
# Copyright (c) 2019 Tomáš Mráz <tmraz@fedoraproject.org>

from subprocess import check_output, CalledProcessError
from tempfile import mkstemp
import os

from .configgenerator import ConfigGenerator


class BindGenerator(ConfigGenerator):
	CONFIG_NAME = 'bind'
	SCOPES = {'dnssec', 'bind'}

	RELOAD_CMD = 'systemctl try-reload-or-restart bind.service 2>/dev/null || :\n'

	sign_not_map = {
		'RSA-MD5':'RSAMD5',
		'DSA-SHA1':'DSA',
		'ECDSA-SHA1':'',
		'RSA-SHA1':'RSASHA1;\nNSEC3RSASHA1'
	}

	hash_not_map = {
		'MD5':'',
		'SHA1':'SHA-1',
		'GOST':'GOST',
		'SHA2-256':'SHA-256',
		'SHA2-384':'SHA-384'
	}

	@classmethod
	def generate_config(cls, policy):
		ip = policy.disabled
		sep = ';\n'

		cfg = 'disable-algorithms "." {\n'
		s = ''
		for i in ip['sign']:
			try:
				s = cls.append(s, cls.sign_not_map[i], sep)
			except KeyError:
				pass
		cfg += cls.append(s, '}', sep)

		cfg = cls.append(cfg, 'disable-ds-digests "." {\n', sep)
		s = ''
		for i in ip['hash']:
			try:
				s = cls.append(s, cls.hash_not_map[i], sep)
			except KeyError:
				pass
		cfg += cls.append(s, '};\n', sep)

		return cfg

	@classmethod
	def test_config(cls, config):
		fd, path = mkstemp()

		try:
			with os.fdopen(fd, 'w') as f:
				f.write('options {\n')
				f.write(config)
				f.write('\n};\n')
			try:
				_ = check_output(["/usr/sbin/named-checkconf", path])
			except CalledProcessError:
				cls.eprint("There is an error in bind generated policy")
				cls.eprint("Policy:\n%s" % config)
				return False
			except OSError:
				# Ignore missing check command
				pass
		finally:
			os.unlink(path)

		return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2019 Red Hat, Inc.
# Copyright (c) 2019 Tomáš Mráz <tmraz@fedoraproject.org>

"""
Lists of algorithms and globbing among them.
"""

import fnmatch

from . import validation


ALL_CIPHERS = (
    'AES-256-GCM', 'AES-256-CCM',
    'AES-192-GCM', 'AES-192-CCM',
    'AES-128-GCM', 'AES-128-CCM',
    'CHACHA20-POLY1305',
    'CAMELLIA-256-GCM', 'CAMELLIA-128-GCM',
    'AES-256-CTR', 'AES-256-CBC',
    'AES-192-CTR', 'AES-192-CBC',
    'AES-128-CTR', 'AES-128-CBC',
    'CAMELLIA-256-CBC', 'CAMELLIA-128-CBC',
    '3DES-CBC', 'DES-CBC', 'RC4-40', 'RC4-128',
    'DES40-CBC', 'RC2-CBC', 'IDEA-CBC', 'SEED-CBC',
    'NULL',
)

ALL_MACS = (
    'AEAD', 'UMAC-128', 'HMAC-SHA1', 'HMAC-SHA2-256',
    'HMAC-SHA2-384', 'HMAC-SHA2-512', 'UMAC-64', 'HMAC-MD5',
)

ALL_HASHES = (
    'SHA2-256', 'SHA2-384', 'SHA2-512', 'SHA3-256', 'SHA3-384', 'SHA3-512',
    'SHA2-224', 'SHA1', 'MD5',
    'GOST',
)

# we disable curves <= 256 bits by default in Fedora
ALL_GROUPS = (
    'X25519', 'SECP256R1', 'SECP384R1', 'SECP521R1', 'X448',
    'FFDHE-1536', 'FFDHE-2048', 'FFDHE-3072', 'FFDHE-4096',
    'FFDHE-6144', 'FFDHE-8192', 'FFDHE-1024',
)

ALL_SIGN = (
    'RSA-MD5', 'RSA-SHA1', 'DSA-SHA1', 'ECDSA-SHA1',
    'RSA-SHA2-224', 'DSA-SHA2-224', 'ECDSA-SHA2-224',
    'RSA-SHA2-256', 'DSA-SHA2-256', 'ECDSA-SHA2-256', 'ECDSA-SHA2-256-FIDO',
    'RSA-SHA2-384', 'DSA-SHA2-384', 'ECDSA-SHA2-384',
    'RSA-SHA2-512', 'DSA-SHA2-512', 'ECDSA-SHA2-512',
    'RSA-SHA3-256', 'DSA-SHA3-256', 'ECDSA-SHA3-256',
    'RSA-SHA3-384', 'DSA-SHA3-384', 'ECDSA-SHA3-384',
    'RSA-SHA3-512', 'DSA-SHA3-512', 'ECDSA-SHA3-512',
    'EDDSA-ED25519', 'EDDSA-ED25519-FIDO', 'EDDSA-ED448',
    'RSA-PSS-SHA1', 'RSA-PSS-SHA2-224', 'RSA-PSS-SHA2-256',
    'RSA-PSS-SHA2-384', 'RSA-PSS-SHA2-512', 'RSA-PSS-RSAE-SHA1',
    'RSA-PSS-RSAE-SHA2-224', 'RSA-PSS-RSAE-SHA2-256',
    'RSA-PSS-RSAE-SHA2-384', 'RSA-PSS-RSAE-SHA2-512',
)


ALL_KEY_EXCHANGES = (
    'PSK', 'DHE-PSK', 'ECDHE-PSK', 'ECDHE', 'RSA',
    'DHE', 'DHE-RSA', 'DHE-DSS', 'EXPORT', 'ANON', 'DH', 'ECDH',
    'DHE-GSS', 'ECDHE-GSS',
)

# Order matters, see preprocess_text
TLS_PROTOCOLS = ('TLS1.3', 'TLS1.2', 'TLS1.1', 'TLS1.0', 'SSL3.0', 'SSL2.0')
DTLS_PROTOCOLS = ('DTLS1.2', 'DTLS1.0', 'DTLS0.9')
IKE_PROTOCOLS = ('IKEv2', 'IKEv1')
ALL_PROTOCOLS = TLS_PROTOCOLS + DTLS_PROTOCOLS + IKE_PROTOCOLS


ALL = {
    'cipher': ALL_CIPHERS,
    'group': ALL_GROUPS,
    'hash': ALL_HASHES,
    'key_exchange': ALL_KEY_EXCHANGES,
    'mac': ALL_MACS,
    'protocol': ALL_PROTOCOLS,
    'sign': ALL_SIGN,
}


def glob(pattern, alg_class):
    """
    Lists algorithms matching a glob, in order of appearance in ALL[alg_class].
    For more examples, refer to tests/unit/parsing/test_alg_lists.py
    >>> glob('RC4-*', 'cipher')
    ['RC4-40', 'RC4-128']
    """
    if alg_class not in ALL:
        raise validation.alg_lists.AlgorithmClassUnknownError(alg_class)

    r = fnmatch.filter(ALL[alg_class], pattern)
    if not r:
        raise validation.alg_lists.AlgorithmEmptyMatchError(pattern, alg_class)
    return r


def earliest_occurrence(needles, ordered_haystack):
    """
    >>> earliest_occurrence('test', 'abcdefghijklmnopqrstuvwxyz')
    'e'
    """
    intersection = [n for n in needles if n in ordered_haystack]
    if not intersection:
        return None
    indices = (ordered_haystack.index(n) for n in intersection)
    return ordered_haystack[min(indices)]


def min_tls_version(versions):
    """
    >>> min_tls_version(['SSL3.0', 'TLS1.2'])
    'SSL3.0'
    """
    return earliest_occurrence(versions, TLS_PROTOCOLS[::-1])


def min_dtls_version(versions):
    """
    >>> min_dtls_version(['DTLS1.2', 'DTLS1.0'])
    'DTLS1.0'
    """
    return earliest_occurrence(versions, DTLS_PROTOCOLS[::-1])


def max_tls_version(versions):
    """
    >>> max_tls_version(['SSL3.0', 'TLS1.2'])
    'TLS1.2'
    """
    return earliest_occurrence(versions, TLS_PROTOCOLS)


def max_dtls_version(versions):
    """
    >>> max_dtls_version(['DTLS1.2', 'DTLS1.0'])
    'DTLS1.2'
    """
    return earliest_occurrence(versions, DTLS_PROTOCOLS)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

import pytest

from python.cryptopolicies.cryptopolicies import ScopeSelector
from python.cryptopolicies.validation.scope import (
    ScopeUnknownError,
    ScopeSelectorEmptyError,
    ScopeSelectorCommaError,
    ScopeSelectorMatchedNothingError,
    ScopeSelectorCurlyBracketsError,
    ScopeSelectorIllegalCharacterError,
)


def test_scope_selector_any():
    scope_any = ScopeSelector()
    assert str(scope_any) == "<ScopeSelector pattern='*'>"
    assert scope_any.matches({'ssh'})
    assert scope_any.matches({'tls', 'gnutls'})
    assert scope_any.matches({})


def test_scope_selector_tls():
    scope_tls = ScopeSelector('tls')
    assert str(scope_tls) == "<ScopeSelector pattern='tls'>"
    assert scope_tls.matches({'tls', 'gnutls'})
    assert not scope_tls.matches({'ssh', 'openssh'})


def test_scope_selector_nontls():
    scope_nontls = ScopeSelector('!tls')
    assert not scope_nontls.matches({'tls', 'gnutls'})
    assert scope_nontls.matches({'ssh', 'openssh'})


def test_scope_selector_posglob():
    scope_posglob = ScopeSelector('tls*')
    assert str(scope_posglob) == "<ScopeSelector pattern='tls*'>"
    assert scope_posglob.matches({'tls'})
    assert not scope_posglob.matches({'gnutls'})
    assert not scope_posglob.matches({'ssh'})
    assert not scope_posglob.matches({'openssh'})


def test_scope_selector_negglob():
    scope_negglob = ScopeSelector('!tls*')
    assert str(scope_negglob) == "<ScopeSelector pattern='!tls*'>"
    assert not scope_negglob.matches({'tls', 'gnutls'})
    assert scope_negglob.matches({'ssh', 'openssh'})


def test_scope_selector_posmixed():
    scope_posmixed = ScopeSelector('{*utls,ssh}')
    assert str(scope_posmixed) == "<ScopeSelector pattern='{*utls,ssh}'>"
    assert scope_posmixed.matches({'tls', 'gnutls'})
    assert scope_posmixed.matches({'ssh', 'openssh'})
    assert not scope_posmixed.matches({'tls', 'openssl'})
    assert not scope_posmixed.matches({'krb5'})


def test_scope_selector_negmixed():
    scope_negmixed = ScopeSelector('!{*utls,ssh}')
    assert str(scope_negmixed) == "<ScopeSelector pattern='!{*utls,ssh}'>"
    assert not scope_negmixed.matches({'tls', 'gnutls'})
    assert not scope_negmixed.matches({'ssh', 'openssh'})
    assert scope_negmixed.matches({'tls', 'openssl'})
    assert scope_negmixed.matches({'krb5'})


def test_scope_selector_curly_brackets():
    for s in ('{', '}', '{a', '{a{', 'a{}', 'a}}', 'a}', '{{},a}'):
        with pytest.raises(ScopeSelectorCurlyBracketsError):
            ScopeSelector(s)


def test_scope_selector_empty():
    for s in ('', '!', '{}', '!{}', '{tls,}'):
        with pytest.raises(ScopeSelectorEmptyError):
            ScopeSelector(s)


def test_scope_selector_illegal_character():
    for s in (' ', '! ', '{ }', '!{ }', '3#', 'a+b'):
        with pytest.raises(ScopeSelectorIllegalCharacterError):
            ScopeSelector(s)


def test_scope_selector_comma():
    with pytest.raises(ScopeSelectorCommaError):
        ScopeSelector(',')
    with pytest.raises(ScopeSelectorCommaError):
        ScopeSelector('tls,ssh')


def test_scope_selector_unknown():
    with pytest.raises(ScopeUnknownError):
        ScopeSelector('nonex')
    with pytest.raises(ScopeUnknownError):
        ScopeSelector('!nonex')
    with pytest.raises(ScopeUnknownError):
        ScopeSelector('!{nonex,tls}')


def test_scope_selector_nomatch():
    with pytest.raises(ScopeSelectorMatchedNothingError):
        ScopeSelector('*nonex*')
    with pytest.raises(ScopeSelectorMatchedNothingError):
        ScopeSelector('!*nonex*')
    with pytest.raises(ScopeSelectorMatchedNothingError):
        ScopeSelector('{tls,*nonex*}')
    with pytest.raises(ScopeSelectorMatchedNothingError):
        ScopeSelector('!{tls,*nonex*}')



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

import textwrap

import pytest

from python.cryptopolicies.cryptopolicies import (
    PolicySyntaxDeprecationWarning,
    preprocess_text,
)


def test_preprocess_text_basics():
    assert (preprocess_text('a=b\nc=d#protocol = TLS1.2\ne=f')
            == 'a = b\nc = d\ne = f')
    assert preprocess_text('a=b\n#protocol = TLS1.2') == 'a = b'
    assert preprocess_text('# commented out protocol = TLS1.2') == ''


def test_preprocess_text_compat():
    with pytest.warns(PolicySyntaxDeprecationWarning):
        # doesn't move or rewrite the rule, just warns
        assert (preprocess_text('protocol = TLS1.2\na=b')
                == 'protocol = TLS1.2\na = b')
    with pytest.warns(PolicySyntaxDeprecationWarning):
        # moves the rule to the end, rewrites it
        assert (preprocess_text('ike_protocol = IKEv1\na=b')
                == 'a = b\nprotocol@IKE = IKEv1')
    with pytest.warns(PolicySyntaxDeprecationWarning):
        assert preprocess_text('''  # reordering is intended
            ike_protocol = y
            protocol = x
            tls_cipher = a b
            cipher = a b c
            ssh_cipher = b c
        ''') == textwrap.dedent('''
            protocol = x
            cipher = a b c
            cipher@TLS = a b
            cipher@SSH = b c
            protocol@IKE = y
        ''').strip()


def test_preprocess_text_compat_problematic():
    with pytest.warns(PolicySyntaxDeprecationWarning):
        # moves the rule to the end, rewrites it
        assert (preprocess_text('tls_cipher = NULL+')
                == 'cipher@TLS = NULL+')
    with pytest.warns(PolicySyntaxDeprecationWarning):
        # moves the rule to the end, rewrites it
        assert (preprocess_text('tls_cipher = \\\n\\\nNULL+\na=b\\\n')
                == 'a = b\ncipher@TLS = NULL+')


def test_preprocess_text_compat_diamond_problem():
    with pytest.warns(PolicySyntaxDeprecationWarning):
        # previous behaviour: at the end of the policy (but not a subpolicy!)
        # derived properties were set to parent properties if missing;
        # e.g., ssh_cipher or tls_cipher were set to cipher

        # the current 'correct' way to disable ciphers, reordering is intended
        assert preprocess_text('''
            # policy
            tls_cipher = a b
            cipher = a b c
            ssh_cipher = b c
        ''') == textwrap.dedent('''
            cipher = a b c
            cipher@TLS = a b
            cipher@SSH = b c
        ''').strip()

        assert preprocess_text('''
            # the current way of disabling ciphers
            ssh_cipher = -b
            cipher = -b
            tls_cipher = -b
        ''') == textwrap.dedent('''
            cipher = -b
            cipher@TLS = -b
            cipher@SSH = -b
        ''').strip()

        # subpolicy which used to affects cipher&ssh_cipher, but not tls_cipher
        assert preprocess_text('''
            # subpolicy
            cipher = -b      # used to not affect tls_cipher (compat insanity)!
            ssh_cipher = -b  # does affect cipher@SSH
        ''') == textwrap.dedent('''
            cipher = -b
            cipher@SSH = -b
        ''').strip()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

import pytest

from python.cryptopolicies.cryptopolicies import Operation, parse_rhs

from python.cryptopolicies.validation.alg_lists import (
    AlgorithmClassUnknownError, AlgorithmEmptyMatchError,
)
from python.cryptopolicies.validation.rules import (
    MixedDifferentialNonDifferentialError,
    IntPropertyNonIntValueError,
    NonIntPropertyIntValueError,
)


def test_parse_rhs():
    assert parse_rhs('+NULL', 'cipher') == [(Operation.PREPEND, 'NULL')]
    assert parse_rhs('-NULL', 'cipher') == [(Operation.OMIT, 'NULL')]
    assert parse_rhs('IDEA-CBC NULL', 'cipher') == [
        (Operation.RESET, None),
        (Operation.APPEND, 'IDEA-CBC'),
        (Operation.APPEND, 'NULL')
    ]
    with pytest.raises(AlgorithmEmptyMatchError):
        parse_rhs('NULL NONEX', 'cipher')
    with pytest.raises(AlgorithmEmptyMatchError):
        parse_rhs('NULL NONEX-*', 'cipher')
    with pytest.raises(AlgorithmClassUnknownError):
        parse_rhs('NULL', 'nonex_algo_class')
    with pytest.raises(MixedDifferentialNonDifferentialError):
        parse_rhs('+IDEA-CBC NULL', 'cipher')

    with pytest.raises(IntPropertyNonIntValueError):
        parse_rhs('something', 'sha1_in_certs')
    with pytest.raises(NonIntPropertyIntValueError):
        parse_rhs('0', 'cipher')
    with pytest.raises(AlgorithmClassUnknownError):
        parse_rhs('0', 'nonex_algo_class')



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

from .general import PolicySyntaxError


class MalformedLine(PolicySyntaxError):
    def __init__(self, line):
        super().__init__(f'malformed line `{line}`')


class MixedDifferentialNonDifferentialError(PolicySyntaxError):
    def __init__(self, rhs):
        super().__init__('cannot initialize list and modify it at once '
                         f'(`{rhs}`)')


class IntPropertyNonIntValueError(PolicySyntaxError):
    def __init__(self, int_property_name):
        # wording follows previous versions
        super().__init__(f'Bad value of policy property `{int_property_name}`:'
                         ' value must be an integer')


class NonIntPropertyIntValueError(PolicySyntaxError):
    def __init__(self, alg_class):
        # wording follows previous versions
        super().__init__(f'Bad value of policy property `{alg_class}`:'
                         ' value must not be an integer')


def count_equals_signs(line):
    if line.count('=') != 1:
        raise MalformedLine(line)


def empty_lhs(lhs, line):
    if not lhs:
        raise MalformedLine(line)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

import pytest

from python.cryptopolicies.alg_lists import (
    glob, ALL,
    min_tls_version, min_dtls_version, max_tls_version, max_dtls_version,
)
from python.cryptopolicies.validation.alg_lists import (
    AlgorithmClassUnknownError, AlgorithmEmptyMatchError,
)


def test_glob_alg_sanity():
    assert glob('SEED-CBC', 'cipher') == ['SEED-CBC']
    assert glob('*EED*', 'cipher') == ['SEED-CBC']
    assert glob('*', 'cipher') == list(ALL['cipher'])


def test_glob_alg_globbing():
    gs = glob('AES-1*', 'cipher')
    assert gs
    assert all(g.startswith('AES-1') for g in gs)


def test_glob_alg_algorithm_empty():
    with pytest.raises(AlgorithmEmptyMatchError):
        glob('NONEX-*', 'cipher')


def test_glob_alg_algorithm_class_unknown():
    with pytest.raises(AlgorithmClassUnknownError):
        glob('*', 'nonex')


def test_min_versions():
    assert min_tls_version(['TLS1.2', 'SSL3.0', 'TLS1.3']) == 'SSL3.0'
    assert min_tls_version(['TLS1.2']) == 'TLS1.2'
    assert min_tls_version(['nonex']) is None
    assert min_tls_version([]) is None
    assert min_dtls_version(['DTLS1.2', 'DTLS1.0']) == 'DTLS1.0'
    assert min_dtls_version(['DTLS1.2']) == 'DTLS1.2'
    assert min_dtls_version(['nonex']) is None
    assert min_dtls_version([]) is None


def test_max_versions():
    assert max_tls_version(['TLS1.2', 'SSL3.0', 'TLS1.3']) == 'TLS1.3'
    assert max_tls_version(['TLS1.2']) == 'TLS1.2'
    assert max_tls_version(['nonex']) is None
    assert max_tls_version([]) is None
    assert max_dtls_version(['DTLS1.2', 'DTLS1.0']) == 'DTLS1.2'
    assert max_dtls_version(['DTLS1.2']) == 'DTLS1.2'
    assert max_dtls_version(['nonex']) is None
    assert max_dtls_version([]) is None



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (c) 2021 Red Hat, Inc.

import fnmatch

from .general import PolicySyntaxError


class ScopeSyntaxError(PolicySyntaxError):
    pass


class ScopeUnknownError(ScopeSyntaxError):
    def __init__(self, scope_glob):
        super().__init__(f'unknown scope {scope_glob}')


class ScopeSelectorEmptyError(ScopeSyntaxError):
    def __init__(self):
        super().__init__('empty scope selector')


class ScopeSelectorIllegalCharacterError(ScopeSyntaxError):
    def __init__(self, selector):
        super().__init__(f'illegal character in scope selector `{selector}`')


class ScopeSelectorCurlyBracketsError(ScopeSyntaxError):
    def __init__(self, pattern):
        super().__init__(f'unsupported curly brackets usage in `{pattern}`')


class ScopeSelectorCommaError(ScopeSyntaxError):
    def __init__(self, pattern):
        super().__init__(f'unsupported comma usage in `{pattern}`')


class ScopeSelectorMatchedNothingError(ScopeSyntaxError):
    def __init__(self, pattern):
        super().__init__(f'scope selector `{pattern}` matches no scope')


def illegal_characters(p, original_pattern):
    if not all(c.isalnum() or c in '{,}*_-' for c in p):
        raise ScopeSelectorIllegalCharacterError(original_pattern)


def curly_brackets(p, original_pattern):
    if ((p.count('{'), p.count('}')) not in [(0, 0), (1, 1)]
            or p.startswith('{') and not p.endswith('}')
            or not p.startswith('{') and p.endswith('}')):
        raise ScopeSelectorCurlyBracketsError(original_pattern)


def resulting_globs(globs, all_scopes, original_pattern):
    if any(',' in g for g in globs):
        raise ScopeSelectorCommaError(original_pattern)
    for g in globs:
        if not g:
            raise ScopeSelectorEmptyError()
        if not fnmatch.filter(all_scopes, g):
            if '*' in g:
                raise ScopeSelectorMatchedNothingError(g)
            raise ScopeUnknownError(g)
